import games
import util


class OthelloGameState(games.GameState):
    """
        Othello game states. 
        
        All you need to know about this class is that it implements the GameState interface,
        and it stores its board in gameState.board as a list of lists of strings.
        In other words, gamesState.board[y][x] = "B", "W", or " " depending on if square (x,y)
        is occupied by a Black piece, White piece, or is empty (respectively).        
    """

    directions = ((0,1), (1,1), (1,0), (1,-1), (0,-1), (-1,-1), (-1,1), (-1, 0))
    actionCalls = 0
    successorCalls = 0
    
    def __init__(self, boardSize = 8, board = None, players = ("B", "W"), previousMove = None, counts = None):
        if board == None: 
            board = [[" ",] * boardSize,] * boardSize
            
            assert boardSize % 2 == 0
            mid1, mid2 = boardSize/2 - 1,boardSize/2
            board[mid1] = board[mid1][:]
            board[mid2] = board[mid2][:]
            board[mid1][mid1] = board[mid2][mid2] = "W"
            board[mid2][mid1] = board[mid1][mid2] = "B"
            counts = (2,2)
        self.boardSize = boardSize
        self.counts = counts
        self.players = players
        self.board = board 
        self.previousMove = previousMove

    def getPlayers(self):
        """  Returns a tuple of player names, where the first player plays next. """
        return self.players


    def getPreviousMove(self):
        """  Returns the most recently made move. """
        return self.previousMove

    def isSquare(self, x, y, player):
        """ Does square (x,y) exist and is occupied by player? """
        return x >= 0 and y >= 0 and x < self.boardSize and y < self.boardSize and self.board[y][x] == player

    def generateActions(self, currentPlayer, otherPlayer):
        anyActions = False
        for y in range(self.boardSize):
            for x in range(self.boardSize):
                if self.board[y][x] == " ":
                    isLegal = False;
                    for (dx, dy) in OthelloGameState.directions:
                        cx, cy = x+dx, y+dy
                        if self.isSquare(cx,cy,otherPlayer):
                            while self.isSquare(cx,cy,otherPlayer):
                                cx, cy = cx+dx, cy+dy
                            if self.isSquare(cx,cy,currentPlayer):
                                isLegal = True
                                break
                    if isLegal:
                        anyActions = True 
                        yield (x,y)
        if not anyActions: yield "Pass"


    def actions(self):
        """  Return a list of actions possible from this state. """
        OthelloGameState.actionCalls += 1
        return [a for a in self.generateActions(*self.players)]
      
        
    def successor(self, action):
        """  Return the successor state of self on action.  
             Precondition: action in actions(self) """
        OthelloGameState.successorCalls += 1
        if action == "Pass":
            return OthelloGameState(self.boardSize, self.board, util.leftRotate(self.players), 
                                    action, util.leftRotate(self.counts))

        (x,y) = action
        assert self.board[y][x] == " "
        (currentPlayer, otherPlayer) = self.players
        
        newBoard = [row[:] for row in self.board]
        (pc, oc) = self.counts
        
        for (dx, dy) in OthelloGameState.directions:
            cx, cy = x+dx, y+dy
            if self.isSquare(cx,cy,otherPlayer):
                while self.isSquare(cx,cy,otherPlayer):
                    cx, cy = cx+dx, cy+dy
                if self.isSquare(cx,cy,currentPlayer):
                    cx, cy = cx-dx, cy-dy
                    while cx != x or cy != y:
                        newBoard[cy][cx] = currentPlayer
                        pc, oc = pc + 1, oc - 1
                        cx, cy = cx-dx, cy-dy
                    newBoard[cy][cx] = currentPlayer
        pc += 1                

        return OthelloGameState(self.boardSize,newBoard, util.leftRotate(self.players), action, (oc, pc))
            
    
    def utility(self):
        """  Returns None if the game is nonterminal, or a Dictionary 
             associating players to utilities if it is terminal. """
        (currentPlayer, otherPlayer) = self.players
        
        if self.generateActions(currentPlayer, otherPlayer).next() == "Pass" and \
           self.generateActions(otherPlayer, currentPlayer).next() == "Pass":
            (pc, oc) = self.counts
            if   pc > oc: pscore = 1
            elif pc < oc: pscore = -1
            else:         pscore = 0
            return {currentPlayer : pscore, otherPlayer : -pscore}
        

    def __str__(self):
        """  Return a printable string representation of this state. """
#        horizontalLine = ("  " + ('-' * (self.boardSize * 4 + 1)))
#        blankLine = "  |" + ("   |" * self.boardSize) 
        horizontalLine = ("  " + ('-' * (self.boardSize * 2 + 1)))
        blankLine = "  |" + (" " * self.boardSize) 

        lines = ["   " + " ".join([str(x) for x in range(self.boardSize)]), horizontalLine]
        for (i, row) in enumerate(self.board):
            lines.append("%d |%s|" % (i, " ".join(row)))
        lines.append(horizontalLine)
        lines.append("Current counts: %s" % zip(self.players, self.counts))
        lines.reverse()
        return '\n'.join(lines)
       

    def representation(self):
        """  Return a hashable representation of self that will be 
             used for representation and equality tests.  """
        return "%s%s" % (self.board, self.players)


    # These methods are used to evaluate SearchAgents    
    def getSearchStats(self):
        """ return (# actions calls, # successor calls) """
        return (OthelloGameState.actionCalls, OthelloGameState.successorCalls)

    
    def resetSearchStats(self):
        """ Reset internal state of problem for tracking search statistics. """    
        OthelloGameState.actionCalls = 0
        OthelloGameState.successorCalls = 0

