import games
import util


class TicTacToeGameState(games.GameState):
    """ 
        TicTacToe game states. 
        
        All you need to know about this class is that it implements the GameState interface.     
    """
    actionCalls = 0
    successorCalls = 0
    
    def __init__(self, board = [[" ",] * 3,] * 3, players = ("X", "O"), previousMove = None):
        self.players = players
        self.board = board 
        self.previousMove = previousMove


    def getPlayers(self):
        """  Returns a tuple of player names, where the first player plays next. """
        return self.players


    def getPreviousMove(self):
        """  Returns the most recently made move. """
        return self.previousMove


    def generateActions(self):
        """  Generate actions possible from this state. """
        actions = []
        for y in range(3):
            for x in range(3):
                if self.board[y][x] == " ":
                    yield (x,y)

    def actions(self):
        """  Return a list of actions possible from this state. """
        TicTacToeGameState.actionCalls += 1
        return [a for a in self.generateActions()]
#    actions = util.memoize(actions)    
        
    def successor(self, action):
        """  Return the successor state of self on action.  
             Precondition: action in actions(self) """
        TicTacToeGameState.successorCalls += 1
        (x,y) = action
        assert self.board[y][x] == " "
        newBoard = self.board[:]
        newBoard[y] = newBoard[y][:]
        newBoard[y][x] = self.currentPlayer()
        return TicTacToeGameState(newBoard, util.leftRotate(self.players), action)
            
    
    def utility(self):
        """  Returns None if the game is nonterminal, or a Dictionary 
             associating players to utilities if it is terminal. """
        currentPlayer = self.players[0]
        lastPlayer    = self.players[1]
        
        # Check for first move
        if not self.previousMove: 
            return None

        # Check for win
        for (xdir, ydir) in ((0,1), (1,1), (1,0), (1,-1)):
            ndir = -1
            for dir in (-1, 1):
                (cx, cy) = self.previousMove
                dx = dir * xdir
                dy = dir * ydir
                while cx >= 0 and cx <= 2 and cy >= 0 and cy <= 2 and self.board[cy][cx] == lastPlayer:
                    ndir += 1
                    cx += dx
                    cy += dy
            if (ndir == 3):
                return {lastPlayer: 1, currentPlayer : -1}
        
        # Check for draw
        try: 
            if self.generateActions().next(): 
                return None
        except StopIteration: 
            return {lastPlayer: 0, currentPlayer : 0} 

       

    def __str__(self):
        """  Return a printable string representation of this state. """
        horizontalLine = ("  " + ('-' * (13)))
        blankLine = "  |   |   |   |" 

        lines = ["    0   1   2  ", horizontalLine]
        for (i, row) in enumerate(self.board):
            lines.append("%d | %s |" % (i, " | ".join(row)))
            lines.append(horizontalLine)
        lines.reverse()
        return '\n'.join(lines)
       

    def representation(self):
        """  Return a hashable representation of self that will be 
             used for representation and equality tests.  """
        return str(self.board)


    # These methods are used to evaluate SearchAgents    
    def getSearchStats(self):
        """ return (# actions calls, # successor calls) """
        return (TicTacToeGameState.actionCalls, TicTacToeGameState.successorCalls)

    
    def resetSearchStats(self):
        """ Reset internal state of problem for tracking search statistics. """    
        TicTacToeGameState.actionCalls = 0
        TicTacToeGameState.successorCalls = 0

